/*
 * Copyright (C) 2015 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { expect } from "chai";
import { Engine } from "./engine";
import { FilterOperator } from "./filter";
import { View } from "./view";
import { engineScenario } from "../../test/scenario";
import { EnumType } from "./enumType";
import { Dimension } from "./dimension";

describe("engine class", () => {

    const engine = new Engine(engineScenario.config);
    const subdim = engineScenario.subDimensions;

    for (let key in subdim){
        if (subdim[key]){
            engine.addDimension(subdim[key]);
        }
    }

    it("should be create a fill that cover all metrics and dimensions", () => {
        let query = engineScenario.queryMetsDims;
        let optimalView = engine.query(query);
        expect(optimalView).to.be.an("object");
        expect(optimalView).to.have.property("metrics");
        expect(optimalView).to.have.property("dimensions");
        expect(optimalView.metrics).to.be.an("array");
        expect(optimalView.dimensions).to.be.an("array");
        expect(optimalView.metrics).to.have.length(16);
        expect(optimalView.dimensions).to.have.length(19);
    });
    it("should throw an exception, query with non-existent metric", () => {
        let error: boolean = false;
        try {
            engine.query(engineScenario.queryNoMets);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal("[Engine Error] Engine views cannot cover the query");

        }
        expect(error);
    });

    it("should throw an exception, query with non-existent dimension", () => {
        let error: boolean = false;
        try {
            engine.query(engineScenario.queryNoDims);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal("[Engine Error] Engine views cannot cover the query");

        }
        expect(error);
    });

    it("should be create a fill that cover the query, that match perfectly with a existent view", () => {
        let query = engineScenario.queryProduct;
        let optimalView = engine.query(query);
        expect(optimalView).to.be.an("object");
        expect(optimalView).to.have.property("metrics");
        expect(optimalView).to.have.property("dimensions");
        expect(optimalView).to.have.property("operation");
        expect(optimalView.metrics).to.be.an("array");
        expect(optimalView.dimensions).to.be.an("array");
        expect(optimalView.operation).to.be.an("object");
        expect(optimalView.metrics).to.have.length(6);
        expect(optimalView.dimensions).to.have.length(3);
        expect(optimalView.operation).to.have.property("opcode");
        expect(optimalView.id).to.be.equal(engineScenario.viewProduct.id);
    });

    it("should be create a fill that cover the query, that match perfectly with a existent view, with clauses", () => {
        let query = engineScenario.queryActive;
        let optimalView = engine.query(query);
        expect(optimalView).to.be.an("object");
        expect(optimalView).to.have.property("metrics");
        expect(optimalView).to.have.property("dimensions");
        expect(optimalView).to.have.property("operation");
        expect(optimalView.metrics).to.be.an("array");
        expect(optimalView.dimensions).to.be.an("array");
        expect(optimalView.operation).to.be.an("object");
        expect(optimalView.metrics).to.have.length(1);
        expect(optimalView.dimensions).to.have.length(2);
        expect(optimalView.operation).to.have.property("opcode");
        expect(optimalView.id).to.be.equal(engineScenario.viewActiveSeller.id);
    });

    it("should be create a fill that cover the query, using sub dimensions", () => {
        let query = engineScenario.querySubDim;
        let optimalView = engine.query(query);
        expect(optimalView).to.be.an("object");
        expect(optimalView).to.have.property("metrics");
        expect(optimalView).to.have.property("dimensions");
        expect(optimalView.metrics).to.be.an("array");
        expect(optimalView.dimensions).to.be.an("array");
        expect(optimalView.metrics).to.have.length(0);
        expect(optimalView.dimensions).to.have.length(2);

        expect(optimalView).satisfy((optView: View) => {
            return optView.dimensions.some((item) => item.name === subdim["subdims_day"].name);
        });
        expect(optimalView).satisfy((optView: View) => {
            return optView.dimensions.some((item) => item.name === subdim["subdims_month"].name);
        });
    });

    it("should throw an exception, sub-dimension with non-existent parent", () => {
        let error: boolean = false;
        try {
            engine.query(engineScenario.queryNoParent);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal("[Engine Error] Engine views cannot cover the query");

        }
        expect(error).to.be.true;
    });

    it("should parse a clause, with target as dimension and operator as ==", () => {
        const strFilter = "dim:client:name==Laci";
        const clause = engine.parseClause(strFilter);
        expect(clause).to.be.an("object");
        expect(clause).to.have.property("filters");
        expect(clause).to.have.property("id");
        expect(clause.filters).to.be.an("array");
        expect(clause.filters).to.have.length(1);
        expect(clause.filters[0]).to.have.property("id");
        expect(clause.filters[0]).to.have.property("target");
        expect(clause.filters[0]).to.have.property("value");
        expect(clause.filters[0]).to.have.property("operator");
        expect(clause.filters[0].target).to.be.equal(engine.getDimensionByName("dim:client:name"));
        expect(clause.filters[0].value).to.be.equal("Laci");
        expect(clause.filters[0].operator).to.be.equal(FilterOperator.EQUAL);
    });

    it("should parse a clause, with target as metric and operator as !=", () => {
        const strFilter = "met:product:avg:pricein!=0";
        const clause = engine.parseClause(strFilter);
        expect(clause).to.be.an("object");
        expect(clause).to.have.property("filters");
        expect(clause).to.have.property("id");
        expect(clause.filters).to.be.an("array");
        expect(clause.filters).to.have.length(1);
        expect(clause.filters[0]).to.have.property("id");
        expect(clause.filters[0]).to.have.property("target");
        expect(clause.filters[0]).to.have.property("operator");
        expect(clause.filters[0]).to.have.property("value");
        expect(clause.filters[0].target).to.be.equal(engine.getMetricByName("met:product:avg:pricein"));
        expect(clause.filters[0].value).to.be.equal("0");
        expect(clause.filters[0].operator).to.be.equal(FilterOperator.NOTEQUAL);
    });

    it("should throw an exception, when a dimension is not found", () => {
        let error: boolean = false;
        const strFilter = "dim:seller:willfail==teste";
        const exeption = "[Engine Error] Filter could not be created: \"dim:seller:willfail\" was not found. Check target spelling.";
        try {
            engine.parseClause(strFilter);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal(exeption);

        }
        expect(error).to.be.true;
    });

    it("should throw an exception, when a metric is not found", () => {
        let error: boolean = false;
        const strFilter = "met:seller:count:willfail==0";
        const exeption = "[Engine Error] Filter could not be created: \"met:seller:count:willfail\" was not found. Check target spelling.";
        try {
            engine.parseClause(strFilter);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal(exeption);

        }
        expect(error).to.be.true;
    });

    it("should throw an exception, when a operator is not found", () => {
        let error: boolean = false;
        let strFilter = "met:sell:sum:quantity=?0";
        let exeption = "[Engine Error] Filter could not be created: Operator on \"" + strFilter + "\" could not be extracted";
        try {
            engine.parseClause(strFilter);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal(exeption);

        }
        expect(error).to.be.true;
        error = false;
        strFilter = "met:sell:sum:quantity!?0";
        exeption = "[Engine Error] Filter could not be created: Operator on \"" + strFilter + "\" could not be extracted";
        try {
            engine.parseClause(strFilter);
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal(exeption);

        }
        expect(error).to.be.true;
    });

    it("should throw an exception, when a operator does not suit", () => {
        const operators = [">", "<", "<=", ">="];
        for (let i = 0; i < operators.length; ++i) {
            let error: boolean = false;
            let strFilter = "dim:client:name" + operators[i] + "Firoz";
            let exeption = "[Engine Error] Filter could not be created: Operator \"" + operators[i] + "\" is invalid for target \"dim:client:name\"";
            try {
                engine.parseClause(strFilter);
            }
            catch (e){
                error = true;
                expect(e.message).to.be.equal(exeption);

            }
            expect(error).to.be.true;
        }
    });

    it("should parse clauses with several operators for dates and integers", () => {
        const operators: {[key: string]: FilterOperator} = {
            ">": FilterOperator.GREATER,
            "<": FilterOperator.LOWER,
            "<=": FilterOperator.LOWEREQ,
            ">=": FilterOperator.GREATEREQ,
            "==": FilterOperator.EQUAL,
            "!=": FilterOperator.NOTEQUAL
        };
        for (let op of Object.keys(operators)) {
            const strFilter = "dim:sell:datein" + op + "2018-02-17";
            const clause = engine.parseClause(strFilter);
            expect(clause).to.be.an("object");
            expect(clause).to.have.property("filters");
            expect(clause).to.have.property("id");
            expect(clause.filters).to.be.an("array");
            expect(clause.filters).to.have.length(1);
            expect(clause.filters[0]).to.have.property("id");
            expect(clause.filters[0]).to.have.property("target");
            expect(clause.filters[0]).to.have.property("operator");
            expect(clause.filters[0]).to.have.property("value");
            expect(clause.filters[0].target).to.be.equal(engine.getDimensionByName("dim:sell:datein"));
            expect(clause.filters[0].value).to.be.equal("2018-02-17");
            expect(clause.filters[0].operator).to.be.equal(operators[op]);
        }

        for (let op of Object.keys(operators)) {
            const strFilter = "dim:seller:id" + op + "0";
            const clause = engine.parseClause(strFilter);
            expect(clause).to.be.an("object");
            expect(clause).to.have.property("filters");
            expect(clause).to.have.property("id");
            expect(clause.filters).to.be.an("array");
            expect(clause.filters).to.have.length(1);
            expect(clause.filters[0]).to.have.property("id");
            expect(clause.filters[0]).to.have.property("target");
            expect(clause.filters[0]).to.have.property("operator");
            expect(clause.filters[0]).to.have.property("value");
            expect(clause.filters[0].target).to.be.equal(engine.getDimensionByName("dim:seller:id"));
            expect(clause.filters[0].value).to.be.equal("0");
            expect(clause.filters[0].operator).to.be.equal(operators[op]);
        }
    });
    it("should return all views", () => {
        let views: View[];
        views = engine.getViews();
        expect(views).to.have.length(9);
        for (let i = 0 ; i < views.length; i++){
            expect(views[i]).to.have.property("metrics");
            expect(views[i]).to.have.property("dimensions");
            expect(views[i]).to.have.property("clauses");
            expect(views[i]).to.have.property("origin");
            expect(views[i]).to.have.property("operation");
            expect(views[i]).to.have.property("id");
            expect(views[i]).to.have.property("name");
        }
    });
    it("should return null from addView", () => {
        let view: View;
        view = engine.addView(null);
        expect(view).to.be.null;
    });
    it("should return null from addDimension", () => {
        let dimension: Dimension;
        dimension = engine.addDimension(null);
        expect(dimension).to.be.null;
    });

    it("should return all views", () => {
        let enumtype: EnumType;
        let error: boolean = false;
        try {
            enumtype = engine.getEnumTypeByName("test");
        }
        catch (e){
            error = true;
            expect(e.message).to.be.equal("[Engine Error] EnumType: 'test' do not exist in the database." +
            " Check enumtype spelling and database configuration.");
        }
        expect(error);
    });

});
