/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { expect } from "chai";

import { Metric } from "../core/metric";
import { Dimension } from "../core/dimension";
import { View } from "../core/view";
import { Filter, FilterOperator } from "../core/filter";
import { Clause } from "../core/clause";
import { Graph } from "./graph";
import { AggregationType, RelationType, DataType } from "../common/types";
import { Query, QueryOpts } from "../common/query";

describe("graph class", () => {

    it("should not create 2 vertices with the same dimension", () => {
        let g = new Graph();
        let dim = new Dimension({name: "dim:test", dataType: DataType.STRING});

        expect(g.addDimension(dim)).to.be.true;
        expect(g.addDimension(dim)).to.be.false;
    });

    it("should not create 2 vertices with same metric", () => {
        let g = new Graph();
        let met = new Metric({
            name: "met:test",
            aggregation: AggregationType.SUM,
            dataType: DataType.STRING
        });

        expect(g.addMetric(met)).to.be.true;
        expect(g.addMetric(met)).to.be.false;
    });

    it("should not create a vertex with a invalid dimension", () => {
        let g = new Graph();
        let dim: Dimension;

        expect(g.addDimension(dim)).to.be.false;
    });

    it("should not create a vertex with a invalid metric", () => {
        let g = new Graph();
        let met: Metric;

        expect(g.addMetric(met)).to.be.false;
    });

    it("should not create a vertex with a subdimension when parent is not a vertex", () => {
        let g = new Graph();
        let dim = new Dimension({name: "dim:test", dataType: DataType.STRING});
        let subdim = new Dimension({
            name: "dim:sub_test",
            dataType: DataType.STRING,
            relation: RelationType.MONTH,
            parent: dim
        });

        expect(g.addDimension(subdim)).to.be.false;
    });

    it("should add a set of views", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.STRING}),
            new Dimension({name: "dim:1", dataType: DataType.STRING}),
            new Dimension({name: "dim:2", dataType: DataType.STRING}),
            new Dimension({name: "dim:3", dataType: DataType.STRING})
        ];

        let g = new Graph();

        for (let i = 0; i < dims.length; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let views = [
            new View({
                metrics: [],
                dimensions: [dims[0], dims[1]],
                origin: true
            }),
            new View({
                metrics: [],
                dimensions: [dims[2], dims[3]],
                origin: true
            }),
            new View({
                metrics: [],
                dimensions: dims,
                origin: true
            })
        ];

        for (let i = 0; i < views.length; ++i) {
            expect(g.addView(views[i])).to.be.true;
        }
    });

    it("should not add a view twice", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.STRING}),
            new Dimension({name: "dim:1", dataType: DataType.STRING}),
        ];

        let g = new Graph();

        for (let i = 0; i < dims.length; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let view = new View({
            metrics: [],
            dimensions: [dims[0], dims[1]],
            origin: true
        });

        expect(g.addView(view)).to.be.true;
        expect(g.addView(view)).to.be.false;
    });

    it("should not add views when metrics and dimensions are not vertices", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.STRING}),
            new Dimension({name: "dim:1", dataType: DataType.STRING}),
            new Dimension({name: "dim:2", dataType: DataType.STRING})
        ];

        let g = new Graph();

        expect(g.addDimension(dims[0])).to.be.true;

        let view = new View({
            metrics: [],
            dimensions: dims,
            origin: true
        });

        expect(g.addView(view)).to.be.false;
    });

    it("should not add invalid views", () => {
        let g = new Graph();
        expect(g.addView(null)).to.be.false;
    });

    it("should add only once a view with only one vertex", () => {
        let dim = new Dimension({name: "dim:0", dataType: DataType.STRING});

        let g = new Graph();

        expect(g.addDimension(dim)).to.be.true;

        let view = new View({
            metrics: [],
            dimensions: [dim],
            origin: true
        });

        expect(g.addView(view)).to.be.true;
        expect(g.addView(view)).to.be.false;
    });

    it("should create a cover for a single vertex", () => {
        let dim = new Dimension({name: "dim:0", dataType: DataType.STRING});

        let g = new Graph();

        expect(g.addDimension(dim)).to.be.true;

        let view = new View({
            metrics: [],
            dimensions: [dim],
            origin: true
        });

        expect(g.addView(view)).to.be.true;
        const qOpts: QueryOpts = {metrics: [], dimensions: [dim]};
        const query = new Query (qOpts);
        let children = g.cover(query);
        expect(children).to.be.an("array");
        expect(children).to.have.length(1);
        expect(children[0].id).to.be.equal(view.id);
    });

    it("should create a cover for several vertices", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.STRING}),
            new Dimension({name: "dim:1", dataType: DataType.STRING}),
            new Dimension({name: "dim:2", dataType: DataType.STRING})
        ];

        let mets = [
            new Metric({name: "met:0", dataType: DataType.INTEGER, aggregation: AggregationType.SUM}),
            new Metric({name: "met:1", dataType: DataType.INTEGER, aggregation: AggregationType.AVG}),
            new Metric({name: "met:2", dataType: DataType.INTEGER, aggregation: AggregationType.AVG})
        ];

        let g = new Graph();

        for (let i = 0; i < 3; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
            expect(g.addMetric(mets[i])).to.be.true;
        }

        let views = [
            new View({
                metrics: [mets[0]],
                dimensions: [dims[0]],
                origin: true
            }),
            new View({
                metrics: [mets[1]],
                dimensions: [dims[1]],
                origin: true
            }),
            new View({
                metrics: [mets[2]],
                dimensions: [dims[2]],
                origin: true
            }),
            new View({
                metrics: [],
                dimensions: dims,
                origin: true
            }),
            new View({
                metrics: mets,
                dimensions: dims,
                origin: true
            }),
            new View({
                metrics: [mets[0], mets[1]],
                dimensions: [dims[0], dims[1]],
                origin: true
            }),
        ];

        for (let i = 0; i < views.length; ++i) {
            expect(g.addView(views[i])).to.be.true;
        }
        const qOpts: QueryOpts = { metrics: [mets[0], mets[1]], dimensions: [dims[0], dims[1]]};
        const query = new Query(qOpts);
        let children = g.cover(query);
        expect(children).to.be.an("array");
        expect(children).to.have.length(1);
        expect(children[0].id).to.be.equal(views[views.length - 1].id);
    });

    it("should create a cover with sub dimensions", () => {
        let dim = new Dimension({name: "dim:0", dataType: DataType.DATE});
        let dims = [
            dim,
            new Dimension({
                name: "subdim:0",
                dataType: DataType.STRING,
                parent: dim,
                relation: RelationType.MONTH
            }),
            new Dimension({
                name: "subdim:1",
                dataType: DataType.STRING,
                parent: dim,
                relation: RelationType.DAY
            }),
        ];

        let g = new Graph();

        for (let i = 0; i < 3; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let view = new View({
            metrics: [],
            dimensions: [dims[0]],
            origin: true
        });

        expect(g.addView(view)).to.be.true;
        const qOpts: QueryOpts = { metrics: [], dimensions: [dims[1], dims[2]]};
        const query = new Query(qOpts);
        let children = g.cover(query);
        expect(children).to.be.an("array");
        expect(children).to.have.length(1);
        expect(children[0].id).to.be.equal(view.id);
    });

    it("should return empty when try to cover a empty list", () => {
        let dim = new Dimension({name: "dim:0", dataType: DataType.DATE});
        let dims = [
            dim,
            new Dimension({
                name: "subdim:0",
                dataType: DataType.STRING,
                parent: dim,
                relation: RelationType.MONTH
            }),
            new Dimension({
                name: "subdim:1",
                dataType: DataType.STRING,
                parent: dim,
                relation: RelationType.DAY
            }),
        ];

        let g = new Graph();

        for (let i = 0; i < 3; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let view = new View({
            metrics: [],
            dimensions: [dims[0]],
            origin: true
        });

        expect(g.addView(view)).to.be.true;
        const qOpts: QueryOpts = { metrics: [], dimensions: []};
        const query = new Query(qOpts);
        let children = g.cover(query);
        expect(children).to.be.an("array");
        expect(children).to.be.empty;
    });

    it("should cover the graph, even when a constraint edge can not be used", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.DATE}),
            new Dimension({name: "dim:1", dataType: DataType.DATE}),
            new Dimension({name: "dim:2", dataType: DataType.DATE}),
        ];

        let filter1 = new Filter({
            target: dims[0],
            operator: FilterOperator.EQUAL,
            value: "01/01/01"
        });
        let filter2 = new Filter({
            target: dims[0],
            operator: FilterOperator.EQUAL,
            value: "01/01/02"
        });
        let clause1 = new Clause({filters: [filter1]});
        let clause2 = new Clause({filters: [filter2]});

        let g = new Graph();

        for (let i = 0; i < 3; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let view1 = new View({
            metrics: [],
            dimensions: [dims[0], dims[1]],
            origin: true
        });

        let view2 = new View({
            metrics: [],
            dimensions: [dims[1], dims[2]],
            origin: true
        });

        let view3 = new View({
            metrics: [],
            dimensions: dims,
            origin: false,
            clauses: [clause1]
        });

        expect(g.addView(view1)).to.be.true;
        expect(g.addView(view2)).to.be.true;
        expect(g.addView(view3)).to.be.true;
        const qOpts: QueryOpts = { metrics: [], dimensions: dims, clauses: [clause2]};
        const query = new Query(qOpts);
        let children = g.cover(query);
        expect(children).to.be.an("array");
        expect(children).to.have.length(2);
        expect(children.every((item) => item.id !==  view3.id)).to.be.true;
    });

    it("should cover the query, using filters of intervals", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.FLOAT}),
            new Dimension({name: "dim:1", dataType: DataType.FLOAT}),
            new Dimension({name: "dim:2", dataType: DataType.DATE}),
        ];

        const filters = [
            new Filter({
                target: dims[0],
                operator: FilterOperator.GREATEREQ,
                value: "0"
            }),
            new Filter({
                target: dims[0],
                operator: FilterOperator.LOWER,
                value: "10"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.GREATER,
                value: "0"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.LOWEREQ,
                value: "10"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.GREATEREQ,
                value: "0"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.LOWER,
                value: "10"
            })
        ];

        const clauses = [
            new Clause({filters: [filters[0]]}),
            new Clause({filters: [filters[1]]}),
            new Clause({filters: [filters[2]]}),
            new Clause({filters: [filters[3]]}),
            new Clause({filters: [filters[4]]}),
            new Clause({filters: [filters[5]]}),
        ];

        const filtersTest = [
            new Filter({
                target: dims[0],
                operator: FilterOperator.GREATER,
                value: "1"
            }),
            new Filter({
                target: dims[0],
                operator: FilterOperator.LOWEREQ,
                value: "9"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.GREATEREQ,
                value: "1"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.LOWER,
                value: "9"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.GREATER,
                value: "1"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.LOWEREQ,
                value: "9"
            })
        ];

        const testClauses = [
            new Clause({filters: [filtersTest[0]]}),
            new Clause({filters: [filtersTest[1]]}),
            new Clause({filters: [filtersTest[2]]}),
            new Clause({filters: [filtersTest[3]]}),
            new Clause({filters: [filtersTest[4]]}),
            new Clause({filters: [filtersTest[5]]}),
        ];

        let g = new Graph();

        for (let i = 0; i < 3; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let view0 = new View({
            metrics: [],
            dimensions: dims,
            origin: false,
            clauses: clauses
        });

        expect(g.addView(view0)).to.be.true;
        const qOpts: QueryOpts = { metrics: [], dimensions: dims, clauses: testClauses};
        const query = new Query(qOpts);
        let children = g.cover(query);
        expect(children).to.have.length(1);
        expect(children[0].id === view0.id).to.be.true;
    });

    it("should cover the query, not using filters of intervals", () => {
        let dims = [
            new Dimension({name: "dim:0", dataType: DataType.FLOAT}),
            new Dimension({name: "dim:1", dataType: DataType.FLOAT}),
            new Dimension({name: "dim:2", dataType: DataType.DATE}),
        ];

        const filters = [
            new Filter({
                target: dims[0],
                operator: FilterOperator.GREATEREQ,
                value: "10"
            }),
            new Filter({
                target: dims[0],
                operator: FilterOperator.LOWER,
                value: "0"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.GREATER,
                value: "10"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.LOWEREQ,
                value: "0"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.GREATEREQ,
                value: "10"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.LOWER,
                value: "0"
            })
        ];

        const clauses = [
            new Clause({filters: [filters[0]]}),
            new Clause({filters: [filters[1]]}),
            new Clause({filters: [filters[2]]}),
            new Clause({filters: [filters[3]]}),
            new Clause({filters: [filters[4]]}),
            new Clause({filters: [filters[5]]}),
        ];

        const filtersTest = [
            new Filter({
                target: dims[0],
                operator: FilterOperator.GREATER,
                value: "0"
            }),
            new Filter({
                target: dims[0],
                operator: FilterOperator.LOWEREQ,
                value: "10"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.GREATEREQ,
                value: "0"
            }),
            new Filter({
                target: dims[1],
                operator: FilterOperator.LOWER,
                value: "11"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.GREATER,
                value: "0"
            }),
            new Filter({
                target: dims[2],
                operator: FilterOperator.LOWEREQ,
                value: "10"
            })
        ];

        const testClauses = [
            new Clause({filters: [filtersTest[0]]}),
            new Clause({filters: [filtersTest[1]]}),
            new Clause({filters: [filtersTest[2]]}),
            new Clause({filters: [filtersTest[3]]}),
            new Clause({filters: [filtersTest[4]]}),
            new Clause({filters: [filtersTest[5]]}),
        ];

        let g = new Graph();

        for (let i = 0; i < 3; ++i) {
            expect(g.addDimension(dims[i])).to.be.true;
        }

        let views = [
            new View({
                metrics: [],
                dimensions: dims,
                origin: false
            })
        ];

        for (let  i = 0; i < testClauses.length; ++i) {
            views.push(new View({
                metrics: [],
                dimensions: dims,
                origin: false,
                clauses: [clauses[i]]
            }));
        }

        for (let  i = 0; i < views.length; ++i) {
            expect(g.addView(views[i])).to.be.true;
        }
        const qOpts: QueryOpts = { metrics: [], dimensions: dims, clauses: testClauses};
        const query = new Query(qOpts);
        let children = g.cover(query);
        expect(children).to.have.length(1);
        expect(children[0].id === views[0].id).to.be.true;
    });

});
